(function (ns) {
  'use strict';

  ns.Cache = function () {
    this.items = {};
  };

  var proto = ns.Cache.prototype;
  
  /**
   * @param {String} name
   * @param {Number} ttl
   * @param {Function} callback
   */
  proto.register = function (name, ttl, callback) {
    this.items[name] = {
      name: name,
      ttl: +ttl,
      callback: callback,

      loading: false,
      time: undefined,
      promise: undefined
    };
    return this;
  };

  proto.has = function (name) {
    return this.items.hasOwnProperty(name);
  };

  /**
   * @param {String} name
   * @returns {Promise}
   */
  proto.get = function (name) {
    var record, dataPromise, resultPromise;
    if (!this.items.hasOwnProperty(name)) {
      throw new Error('Cache: item with name "' + name + '" is not registered');
    }

    record = this.items[name];
    if (null == record.time || (record.ttl > 0 && Date.now() > record.time + record.ttl)) {
      if (record.loading && record.promise) {
        return record.promise;
      }

      record.loading = true;
      record.promise = new Promise(function (resolve, reject) {
        record.callback(resolve, reject);
      });
      record.promise.then(function () {
        record.loading = false;
        record.time = Date.now();
      }, function () {
        record.loading = false;
        record.time = Date.now();
      });
    }

    return record.promise;
  };

  ns.cache = new ns.Cache();

})(o35.defineNS('app'));
