(function (ns, $) {

  // todo: use cr.setTemplateBind('CommentText', function (){}) to watch changes in comments

  var arraySlice = Array.prototype.slice;

  ns.CopyComment = function () {
    this.construct.apply(this, arguments);
  };
  var proto = ns.CopyComment.prototype;

  proto.construct = function () {
    this.registerDelegateListener();
    this.injectLinks(document.documentElement);
  };

  proto.registerDelegateListener = function () {
    document.addEventListener('click', this.onBtnClick.bind(this));
  };

  proto.onBtnClick = function (e) {
      if (!e.target.matches('.o35.btn-copy-html')) {
        return;
      }
      e.preventDefault();    

      var node = e.target;
      while ((node = node.parentNode) && !node.matches('.comment-row')) {}
      var commentId = node.getAttribute('_id');
      this.copyComment(commentId);
  };

  proto.copyComment = function (commentId) {
    var self = this;
    var issueId = document.querySelector('.issueId').textContent;
    this.getCommentText(issueId, commentId, function (err, text) {
      if (err) {
        o35.debug && console.error('O35 CopyComment', err);
        return;
      }
      text = self.convertComment(text);
      self.renderComment(text, function (err, html) {
        if (err) {
          o35.debug && console.error('O35 CopyComment', err);
          return;
        }
        o35.setClipboard('<div>' + html + '</div>');
      });
    });
  };

  proto.getCommentText = function (issueId, commentId, callback) {
    app.getIssue(issueId).then(function (xml) {
      var comment = xml.getElementById(commentId);
      comment = comment.getAttribute('text');
      callback(null, comment);
    }, function (err) {
      callback(err instanceof Error ? err : new Error(err));
    });
  };

  proto.convertComment = function (text) {
    var preservedList = [];
    var preserve = function (fragment) {
      var uid = preservedList.length;
      preservedList.push(fragment);
      return '%{' + uid + '}%';
    };

    text = text.replace(/\{html\}([\s\S]*?)\{html\}/gi, function (match, text) {
      return preserve(text);
    });
    text = text.replace(/\{monospace\}([\s\S]*?)\{monospace\}/gi, function (match, text) {
      return preserve('`' + text + '`');
    });
    text = text.replace(/\{noformat\}([\s\S]*?)\{noformat\}/gi, function (match, text) {
      return preserve('<pre>\n' + text + '\n</pre>');
    });
    // fix headers
    text = text.replace(/^(=+)([^=\n]+)\1/gim, function (match, symbols, text) {
      return preserve(new Array(symbols.length + 1).join('#') + text);
    });
    text = text.replace(/\{color:\s*(.*?)\s*\}([\s\S]*?)\{color\}/gi, function (match, color, text) {
      return preserve('<span style="color: ' + color + '">' + text + '</span>');
    });
    text = text.replace(/\{code([^\}\n]*?)\}([\s\S]*?)\{code\}/gi, function (match, attr, text) {
      var title, lang, attrMatch;
      if (attr && (attrMatch = attr.match(/title=(.+)/))) {
        title = attrMatch[1];
      }
      if (attr && (attrMatch = attr.match(/lang=(\w+)/))) {
        lang = attrMatch[1];
      }
      return preserve((title ? title + '\n' : '') + '```' + (lang ? lang : '') + '\n' + text + '\n```');
    });

    // fix tables
    text = text.replace(/\n(\|.+\|)\n(\|.+\|)\n/gi, function (match, head, line) {
      var headMatch;
      head = head.replace(/\|\|/gi, '|').trim();
      line = line.replace(/\|\|/gi, '|').trim();
      headMatch = head.match(/\|/gi);
      if (headMatch) {
        head += '\n' + headMatch.join('-');
      }
      return preserve('\n' + head + '\n' + line + '\n');
    });

    // fix emphasis
    text = text.replace(/('{2,})(.+)\1/gim, function (match, mod, text) {
      var len = mod.trim().length;
      text = text.trim();
      // strong
      if (len >= 3) {
        len -= 3;
        text = '**' + text + '**';
      }
      // italics
      if (len >= 2) {
        len -= 2;
        text = '*' + text + '*';
      }
      return preserve(text);
    });
    text = text.replace(/--([^-\n]+)--/gim, function (match, text) {
      text = text.trim();
      return preserve('~~' + text + '~~');
    });
    text = text.replace(/\*([^*\n ][^*\n]+)\*/gim, function (match, text) {
      text = text.trim();
      return preserve('**' + text + '**');
    });

    // fix lists
    text = text.replace(/^[#\*-](.*)/gim, function (match, text) {
      return preserve('*' + text);
    });

    // [query:]
    text = text.replace(/\[query:([^\]\n]+)\]/gim, function (match, text) {
      return preserve('https://my.orange35.com/youtrack/issues?q=' + encodeURIComponent(text.trim()));
    });

    // @user
    text = text.replace(/(\s)@(\w+)\b/gim, function (match, space, text) {
      return space + preserve('https://my.orange35.com/youtrack/user/' + encodeURIComponent(text.trim()));
    });

    // -----
    text = text.replace(/-{5,}/gim, function (match) {
      return preserve('\n---\n');
    });

    // links
    text = text.replace(/\[(\w+:\/\/[^ \n]+)([^\]]+)\]/, function (match, link, text) {
      return preserve('[' + text.trim() + '](' + link.trim() + ')');
    });

    // finish
    text = text.replace(/%{(\d+)}%/gi, function (match, code) {
      code = +code;
      return code < preservedList.length ? preservedList[code] : '<!-- ' + code + ' -->';
    });

    return text;
  };

  proto.renderComment = function (text, callback) {
    marked(text, {breaks: true}, callback);
  };

  proto.injectLinks = function (container) {
    var btn = this.createLink();
    var issueList = container.querySelector('.issue-comments');
    arraySlice.call(issueList.querySelectorAll('.comment-actions')).forEach(function (node) {
      node.parentNode.appendChild(btn.cloneNode(true));
    });
  };

  proto.createLink = function () {
    var btn = document.createElement('a');
    btn.className = 'o35 btn-copy-html';
    btn.setAttribute('href', '#');
    btn.setAttribute('data-o35-btn', 'copy-html');
    btn.textContent = 'Copy HTML';
    return btn;
  };

  proto.injectStylesheet = function (href) {
    var link = document.createElement('link');
    link.rel = 'stylesheet';
    link.type = 'text/css';
    link.href = href;
    link.media = 'all';
    document.head.appendChild(link);
  };

  proto.injectScript = function (src, callback) {
    var link = document.createElement('script');
    link.type = 'text/javascript';
    link.onload = function () { callback(null); };
    link.onerror = function () { callback(new Error('Can not load script "' + src + '"')); };
    link.src = src;
    document.head.appendChild(link);
  };

})(o35.defineNS('app.issue'), Zepto);
