(function(ns, $) {

  var arraySlice = Array.prototype.slice;

  var propertyMap = {
    'dueDate': 'Due Date',
    'assignee': 'Assignee',
    'htmlCoder': 'HTML Coder',
    'reviewer': 'Reviewer',
    'tester': 'Tester',
    'spentTime': 'Spent time',
    'estimation': 'Estimation',
    'subsystem': 'Subsystem',
    'estimationDevelopment': 'Estimation Development',
    'estimationHTMLing': 'Estimation HTMLing',
    'estimationReview': 'Estimation Review',
    'estimationTesting': 'Estimation Testing',
    'timerTime': 'Timer time',
    'htmlingTimer': 'HTMLing timer',
    'reviewTimer': 'Review timer',
    'testingTimer': 'Testing timer'
  };

  var templates = {
    progressTitle: 'Progress: {progressPercent}% ({spentHrs}h / {estimatedHrs}h)',
    overlimitTitle: 'Overlimit: {overlimitBarPercent}% ({spentHrs}h / {estimatedHrs}h; over: {overtimeHrs}h)',
    
    progressbar: [
      '<div data-o35-progress class="o35-time-progress" style="position: relative; width: 100%; height: 100%; background-color: #dfdfdf; line-height: 5px;">',
      '    <div style="position: absolute; width: {progressPercent}%; background-color: #0099ff;">&nbsp;</div>',
      '    <div style="position: absolute; width: {overlimitPercent}%; background-color: #dd0033;">&nbsp;</div>',
      '    <div style="position: absolute; width: 100%; text-align: center; color: #353535;" title="{title}">&nbsp;</div>',
      '    &nbsp;',
      '</div>'
    ].join('\n')
  };
    
  ns.Sidebar = function () {
    this.construct.apply(this, arguments);
  };
  var proto = ns.Sidebar.prototype;

  proto.construct = function () {
    o35.debug && console.info('O35: app.issue.Sidebar');
    this.infoLinksInitiated = false;
    this.initVerticalProperties();
    $(window).on('tplVerticalPropertiesChange', $.proxy(this.initVerticalProperties, this));
    (function () {
      var s = document.createElement('script');
      s.async = true;
      s.textContent = [
        "cr.setTemplateBind('VerticalProperties', function () {",
        "window.dispatchEvent(new Event('tplVerticalPropertiesChange'));",
        "});"
      ].join('\n');
      document.documentElement.appendChild(s);
    })();
  };

  proto.initVerticalProperties = function () {
    o35.debug && console.info('O35: initiate timer progress bars');
    var issueId = app.getIssueId();
    var issueData, subsystem;
    var projectId = app.getProjectId();

    Promise.all([
      app.getIssue(issueId),
      app.getIssueWorkItems(issueId)
    ]).then((function (values) {
      var issueXmlData = values[0], workItems = values[1];
      issueData = this.getIssueData(issueXmlData);
      subsystem = issueData[propertyMap.subsystem];
      this.render(issueData, workItems);
      o35.debug && console.info('O35: initiate info links');
      app.getProjectInfo(projectId, subsystem).then(
          (function (value) {
            this.insertInfoLinks(value);
          }).bind(this),
          function (err) {
            o35.debug && console.error('O35 Issue', err);
        }
      );
    }).bind(this), function (err) {
      o35.debug && console.error('O35 Issue', err);
    });

  };

  /**
   * @param {Object} info
   * @param {String} info.id
   * @param {String} info.title
   * @param {String} info.projectManager
   * @param {String} info.leadDeveloper
   * @param {String} info.htmlCoder
   * @param {String} info.tester
   * @param {String} info.repositoryName
   * @param {String} info.infoUrl
   * @param {String} info.wikiProjectPage
   * @param {String} info.wikiAccessPage
   */
  proto.insertInfoLinks = function (info) {
    if (document.querySelector('[data-o35-links]')) {
      return;
    }
    var container = $('.fsi-properties');
    var row = container.find('tr:first');
    var links = [], repositoryUrl, repositoryText = 'vcs';
    info.infoUrl && links.push('<a href="' + o35.escapeHtml(o35.prepareLink(info.infoUrl)) + '" target="_blank">info</a>');
    info.wikiProjectPage && links.push('<a href="' + o35.escapeHtml(o35.prepareLink(info.wikiProjectPage, o35.wikiBaseUrl)) + '" target="_blank">project</a>');
    info.wikiAccessPage && links.push('<a href="' + o35.escapeHtml(o35.prepareLink(info.wikiAccessPage, o35.wikiBaseUrl)) + '" target="_blank">access</a>');
    if (info.repositoryName) {
      if (-1 === info.repositoryName.indexOf('http')) {
        repositoryUrl = 'https://my.orange35.com/WebSVN/listing.php?repname=' + info.repositoryName;
      } else {
        repositoryUrl = info.repositoryName;
      }
      if (0 === repositoryUrl.indexOf('https://my.orange35.com/WebSVN')) {
        repositoryText = 'svn';
      }
      if (0 === repositoryUrl.indexOf('https://git.orange35.com')
        || 0 === repositoryUrl.indexOf('https://github.com/')
      ) {
        repositoryText = 'git';
      }
    }
    repositoryUrl && links.push('<a href="' + repositoryUrl + '" target="_blank">' + repositoryText+ '</a>');
    row.after('<tr data-o35-links class="o35-vp-links">' + (links.length ? '<td>Links</td><td>(' + links.join(', ') + ')</td>' : '') + '</tr>');

    if (info.projectManager) {
      row.after('<tr class="o35-usr-pm"><td>PM</td><td>' + o35.escapeHtml(info.projectManager) + '</td></tr>');
    }
    if (info.leadDeveloper) {
      row.after('<tr class="o35-usr-pm"><td>LD</td><td>' + o35.escapeHtml(info.leadDeveloper) + '</td></tr>');
    }
  };

  proto.render = function (issueData, workItems) {
    var name, item,
      timeItems = this.parseWorkTime(workItems),
      userTime = this.getUserTime(timeItems),
      userDailyTime = this.getUserDailyTime(timeItems),
      properties = {},
      residual = +issueData[propertyMap.spentTime];

    console.log('issueData', issueData);
    this.renderTimeSpentBlock(userTime, userDailyTime);

    properties[propertyMap.estimation] = { prop: propertyMap.spentTime };
    properties[propertyMap.estimationDevelopment] = { user: issueData[propertyMap.assignee] };
    properties[propertyMap.estimationHTMLing] = { user: issueData[propertyMap.htmlCoder] };
    properties[propertyMap.estimationReview] = { user: issueData[propertyMap.reviewer] };
    properties[propertyMap.estimationTesting] = { user: issueData[propertyMap.tester] };

    // find holders
    var rows = document.querySelector('.fsi-properties').querySelectorAll('.fsi-property-label');
    var rowsToHide = {};
    rowsToHide[propertyMap.timerTime] = 1;
    rowsToHide[propertyMap.htmlingTimer] = 1;
    rowsToHide[propertyMap.reviewTimer] = 1;
    rowsToHide[propertyMap.testingTimer] = 1;
    
    arraySlice.call(rows).forEach(function (node) {
      var label = String(node.textContent).trim();
      if (properties.hasOwnProperty(label)) {
        properties[label].node = $(node).closest('tr').find('.fsi-property').get(0);
      }
      if (rowsToHide.hasOwnProperty(label)) {
        $(node).closest('tr').hide();
      }
    });

    // calculate time and fix development (development = all - (review + design + testing))
    for (name in properties) {
      if (properties.hasOwnProperty(name) && name !== propertyMap.estimationDevelopment) {
        item = properties[name];
        spent = ('user' in item) ? userTime[item.user] : 0;
        residual -= spent ? spent : 0;
      }
    }
    userTime[properties[propertyMap.estimationDevelopment].user] = residual;

    // calculate progress and overlimit
    for (name in properties) {
      if (properties.hasOwnProperty(name)) {
        item = properties[name];
        var estimate, spent;
        item.progress = item.overlimit = item.overtime = null;
        estimate = issueData[name];
        spent = 'user' in item 
          ? userTime[item.user]
          : ('prop' in item ? issueData[item.prop] : 0);
        if ((item.user || item.prop) && null != estimate && null != spent) {
          item.progress = spent / estimate;
          item.estimate = +estimate;
          item.spent = +spent;
          item.overtime = spent - estimate;
          if (item.progress > 1) {
            item.overlimit = item.progress - 1;
            item.progress = 1;
          }
        }
      }
    }

    // render progressbars
    var templateData, template;
    var renderTemplate = function (template, data) {
      return template.replace(/\{([\w\d_-]+)\}/g, function (match, placeholder) {
        return data.hasOwnProperty(placeholder) ? data[placeholder] : '{' + placeholder + '}';
      });
    };

    for (name in properties) {
      if (properties.hasOwnProperty(name)) {
        item = properties[name];
        if (item.node) {
          templateData = {
            progressPercent: (item.progress * 100).toFixed(2),
            overlimitPercent: Math.min(100, item.overlimit * 100).toFixed(2),
            overlimitBarPercent: (item.overlimit * 100).toFixed(2),
            spentHrs: item.spent ? (item.spent / 60).toFixed(2) : '-',
            estimatedHrs: item.estimate ? (item.estimate / 60).toFixed(2) : '-',
            overtimeHrs: item.overtime ? (item.overtime / 60).toFixed(2) : '-'
          };

          if (+item.overlimit > 0) {
            templateData.title = renderTemplate(templates.overlimitTitle, templateData);
          } else {
            templateData.title = renderTemplate(templates.progressTitle, templateData);
          }         
          template = renderTemplate(templates.progressbar, templateData);

          $(item.node)
            .find('[data-o35-progress]').remove()
            .andSelf().append(template)
            .find('.attribute.cf-wrapper:first').attr('title', templateData.title);
              
        }
      }
    }
    return this;
  };

  proto.renderTimeSpentBlock = function (userTime, dailyTime) {
    $('[data-spent-time]').remove();
    if ($.isEmptyObject(userTime)) {
      return;
    }
    var table = $('<table data-spent-time class="spent-time">');
    $('.fsi-properties-container').append(table);
    var hours, i, dates = [], countDates;
    for (var date in dailyTime) {
      if (dailyTime.hasOwnProperty(date)) {
        dates.push(date);
      }
    }
    dates.sort(function (a, b) { return a > b});
    countDates = dates.length > 3 ? 3 : dates.length;
    table.find('[data-o35-user-time]').remove();
    table.append($('<tr>').append('<td class="title" colspan="' + (2 + countDates) + '">Time Spent</td>'));

    var headers = $('<tr>').append('<th>User</th><th>Spent</th>');
    for (i = 0; i < dates.length && i < 3; i++) {
      date = new Date(dates[i]);
      headers.append('<th><span title="' + this.formatDate(date, 'YYYY-MM-dd') +'">' + this.formatDate(date, 'MMM d') + '</span></th>');
    }
    table.append(headers);


    for (var name in userTime) {
      if (!userTime.hasOwnProperty(name)) {
        continue;
      }
      var row = $('<tr data-o35-user-time>').append('<td>' + name + '</td>');
      row.append('<td style="padding-left: 10px">' + this.toFriendlyTime(userTime[name]) + '</td>');
      for (i = 0; i < dates.length && i < 3; i++) {
        date = dates[i];
        var time = dailyTime[date][name] || '';
        if (time) {
          time = this.toFriendlyTime(time);
        }
        row.append('<td>' + time + '</td>')
      }
      table.append(row);
    }
  };


  proto.toFriendlyTime = function (minutes) {
    var h = Math.floor(minutes / 60);
    var m = minutes % 60;
    var time = '';
    if (h > 0) {
      time += '' + h + 'h';
    }
    if (m > 0) {
      time += '' + m + 'm';
    }
    return time;
  };


  /**
   * @param {DOMParser} issue
   * @returns {Object}
   */
  proto.getIssueData = function (issue) {
    var issueData = {};
    [
      propertyMap.dueDate,
      propertyMap.assignee,
      propertyMap.htmlCoder,
      propertyMap.reviewer,
      propertyMap.tester,
      propertyMap.spentTime,
      propertyMap.estimation,
      propertyMap.estimationDevelopment,
      propertyMap.estimationHTMLing,
      propertyMap.estimationReview,
      propertyMap.estimationTesting,
      propertyMap.subsystem
    ].forEach(function (name) {
      var node = issue.querySelector('field[name="' + name + '"] > value');
      issueData[name] = node ? node.textContent : null;
    });
    return issueData;
  };

  /**
   * @param {DOMParser} workItems
   * @returns {Object}
   */
  proto.parseWorkTime = function (workItems) {
    var result = [];
    var list = workItems.querySelectorAll('workItem');
    arraySlice.call(list).forEach((function (node) {
      var duration = node.querySelector('duration').textContent,
        date = new Date(parseInt(node.querySelector('date').textContent, 10)),
        author = node.querySelector('author').getAttribute('login');
      result.push({
        date: this.formatDate(date, 'YYYY-MM-dd'),
        author: author,
        duration: parseInt(duration, 10)
      });
    }).bind(this));
    return result;
  };

  proto.formatDate = function (dt, format) {
    if (typeof dt === 'string'){
      dt = new Date(dt);
    }
    var pad = function(n, width, z) {
      z = z || '0';
      n = n + '';
      return n.length >= width ? n : new Array(width - n.length + 1).join(z) + n;
    };
    var monthNames = [
      "January", "February", "March",
      "April", "May", "June", "July",
      "August", "September", "October",
      "November", "December"
    ];

    var Days = [
      "Sunday", "Monday", "Tuesday", "Wednesday",
      "Thursday", "Friday", "Saturday"
    ];
    format = format.replace('ss', pad(dt.getSeconds(),2));
    format = format.replace('s', dt.getSeconds());
    format = format.replace('dd', pad(dt.getDate(),2));
    format = format.replace('d', dt.getDate());
    format = format.replace('mm', pad(dt.getMinutes(),2));
    format = format.replace('m', dt.getMinutes());
    format = format.replace('MMMM', monthNames[dt.getMonth()]);
    format = format.replace('MMM', monthNames[dt.getMonth()].substring(0,3));
    format = format.replace('MM', pad(dt.getMonth()+1,2));
    format = format.replace(/M(?![ao])/, dt.getMonth()+1);
    format = format.replace('DD', Days[dt.getDay()]);
    format = format.replace(/D(?!e)/, Days[dt.getDay()].substring(0,3));
    format = format.replace('yyyy', dt.getFullYear());
    format = format.replace('YYYY', dt.getFullYear());
    format = format.replace('yy', (dt.getFullYear()+"").substring(2));
    format = format.replace('YY', (dt.getFullYear()+"").substring(2));
    format = format.replace('HH', pad(dt.getHours(),2));
    format = format.replace('H', dt.getHours());
    return format;
  };

  proto.getUserTime = function (timeItems) {
    var result = {};
    timeItems.forEach(function (item) {
      if (!result.hasOwnProperty(item.author)) {
        result[item.author] = 0.0;
      }
      result[item.author] += item.duration;
    });
    return result;
  };

  proto.getUserDailyTime = function (timeItems) {
    var result = {};
    timeItems.forEach(function (item) {
      if (!result.hasOwnProperty(item.date)) {
        result[item.date] = {};
      }
      if (!result[item.date].hasOwnProperty(item.author)) {
        result[item.date][item.author] = 0;
      }
      result[item.date][item.author] += item.duration;
    });
    return result;
  }
})(o35.defineNS('app.issue'), Zepto);
