(function (o35, app, $) {
  'use strict';

  /**
   * @param {String} issueId
   * @returns {DOMParser}
   */
  app.getIssue = function (issueId) {
    var name = 'issue-' + issueId;
    if (!app.cache.has(name)) {
      app.cache.register(name, 10000, function (resolve, reject) {
        $.ajax({
          type: 'GET',
          url: o35.youtrackUrl + '/rest/issue/' + issueId,
          success: function (data, msg, xhr) {
            var parser = new window.DOMParser();
            var xml = parser.parseFromString(xhr.responseText, "text/xml");
            resolve(xml);
          },
          error: function (xhr, type) {
            reject(new Error(xhr.statusText + '\n' + xhr.responseText));
          }
        });
      });
    }

    return app.cache.get(name);
  };

  /**
   * @param {String} issueId
   * @returns {DOMParser}
   */
  app.getIssueWorkItems = function (issueId) {
    var name = 'issue-workItem-' + issueId;
    if (!app.cache.has(name)) {
      app.cache.register(name, 10000, function (resolve, reject) {
        $.ajax({
          type: 'GET',
          url: o35.youtrackUrl + '/rest/issue/' + issueId + '/timetracking/workitem',
          success: function (data, msg, xhr) {
            var parser = new window.DOMParser();
            var xml = parser.parseFromString(xhr.responseText, "text/xml");
            resolve(xml);
          },
          error: function (xhr, type) {
            reject(new Error(xhr.statusText + '\n' + xhr.responseText));
          }
        });
      });
    }

    return app.cache.get(name);
  };

  app.getProjectInfo = function (projectId, subsystem) {
    var name = 'project-' + projectId;
    if (!app.cache.has(name)) {
      app.cache.register(name, 60000, function (resolve, reject) {
        var apiUrl = o35.baseUrl + '/api/project/' + projectId;
        if (subsystem) {
          apiUrl += '?subsystem=' + subsystem;
        }
        $.ajax({
            method: 'GET',
            url: apiUrl,
            success: function (data, status, xhr) {
              var info;
              if (200 !== xhr.status) {
                o35.debug && console.error('O35 Error: can not get project info', apiUrl, xhr);
                reject(new Error('O35 Error: can not get project "' + projectId + '" info'));
                return;
              }
              try {
                info = JSON.parse(xhr.responseText);
              } catch (e) {
                o35.debug && console.error('O35 Error: can not parse project info response', xhr.responseText);
                reject(new Error('O35 Error: can not parse project info response'));
                return;
              }

              resolve(info.data);
            },
            error: function (xhr) {
              reject(new Error(xhr.statusText + '\n' + xhr.responseText));
            }
        });
      });
    }

    return app.cache.get(name);
  };

  app.getIssueId = function () {
    var node = document.querySelector('.issueId'), issueId;
    if (null == node || '' == (issueId = node.textContent)) {
      throw new Error('O35: Can not get issueId');
    }
    return issueId;
  };

  app.getProjectId = function () {
    var issueId = app.getIssueId();
    return issueId.substring(0, issueId.indexOf('-'));
  };

  app.getSubsystem = function () {
    return this.subsystem;
  };

  app.setSubsystem = function (subsystem) {
    this.subsystem = subsystem;
  };


  $(function () {
    new app.issue.CopyComment();
    new app.issue.Detach();
    new app.issue.Sidebar();
  });

})(o35, o35.defineNS('app'), Zepto);